package masterelection

import (
	"log"
	"time"

	"github.com/coreos/go-etcd/etcd"
)

const (
	STATE_SLAVE = iota
	STATE_MASTER
)

type MasterElection struct {
	client *etcd.Client
	stop chan bool
	stopped bool

	Addr string
	MasterAddr string
	Path string
	TTL  uint64

	State       int
	StateChange chan int
}

func NewMasterElection(client *etcd.Client, path, addr string, ttl uint64, sch chan int, stop chan bool) *MasterElection {
	if ttl < 2 {
		ttl = 2
	}
	return &MasterElection{
		client:      client,
		Path:        path,
		Addr:        addr,
		TTL:         ttl,
		State:       STATE_SLAVE,
		StateChange: sch,
		stop:        stop,
	}
}

func (m *MasterElection) IsMaster() bool {
	return m.State == STATE_MASTER
}

func (m *MasterElection) setState(state int) {
	if m.State == state {
		return
	}
	if m.StateChange != nil {
		m.StateChange <- state
	}
	m.State = state
}

func (m *MasterElection) stopper() {
	<-m.stop

	// Tell the Run() goroutine to exit as soon as it can (we
	// could have simply used the 'stop' channel there but Watch()
	// ignores the stop channel if invoked without a receiving
	// channel and we'd have to refactor Run() to use a temp
	// channel for each invocation.
	m.stopped = true

	// Remove the lock file if we are the master.
	if m.State == STATE_MASTER {
		m.client.Delete(m.Path)
	}
}

func (m *MasterElection) Run() {
	go m.stopper()

	// Start as a slave.
	m.setState(STATE_SLAVE)

	halfttl := time.Second * time.Duration(m.TTL / 2)

	for !m.stopped {
		// Try to acquire the lock. If we are currently the
		// master, the previous value should be our own
		// address, otherwise it should be unset.
		prevValue := ""
		if m.State == STATE_MASTER {
			prevValue = m.Addr
		}
		resp, ok, err := m.client.TestAndSet(m.Path, prevValue, m.Addr, m.TTL)
		if err != nil {
			log.Printf("%s: error from etcd: %s", m.Path, err)
			time.Sleep(20 * time.Millisecond)
			continue
		}

		if ok {
			// Howdy, we're the master now. Wait a while
			// and renew our TTL.
			m.setState(STATE_MASTER)
			m.MasterAddr = m.Addr
			time.Sleep(halfttl)
		} else {
			// We're not the master. Watch for a DELETE
			// (in theory, but we're not actually
			// verifying the action type, just waiting for
			// the first event...)
			m.setState(STATE_SLAVE)
			m.MasterAddr = resp.PrevValue
			_, err := m.client.Watch(m.Path, resp.Index, nil, nil)
			if err != nil {
				log.Printf("%s: watch error: %s", m.Path, err)
			}
		}
	}
}
